"""
Run Hoppr processing, using multiple processors
"""
from __future__ import annotations

import logging
import os
import pwd
import shutil
import socket
import tempfile
import urllib
from concurrent.futures import Future, ProcessPoolExecutor
from datetime import datetime
from itertools import zip_longest
from pathlib import Path
from typing import Dict, List, Optional, Tuple

from packageurl import PackageURL  # type: ignore
from typer import echo

from hoppr.base_plugins.hoppr import HopprPlugin
from hoppr.configs.credentials import Credentials
from hoppr.configs.manifest import Manifest
from hoppr.configs.transfer import Transfer
from hoppr.context import Context
from hoppr.exceptions import HopprPluginError
from hoppr.flatten_sboms import flatten_sboms
from hoppr.result import Result
from hoppr.types.cyclonedx_1_4 import Component
from hoppr.types.transfer_file_content import Plugin as PluginRef
from hoppr.utils import download_file, plugin_instance


def _run_plugin(
    plugin: HopprPlugin, method_name: str, component: Optional[Component]
) -> Result:
    """
    Runs a single method for a single component (if supplied) on a single plugin
    """
    if method_name == HopprPlugin.pre_stage_process.__name__:
        result = plugin.pre_stage_process()
    elif method_name == HopprPlugin.process_component.__name__:
        result = plugin.process_component(component)
    elif method_name == HopprPlugin.post_stage_process.__name__:
        result = plugin.post_stage_process()
    else:
        result = Result.fail(f"Invalid method: {method_name}")

    return result


class StageProcessor:  # pylint: disable=too-few-public-methods
    """
    Class to handle all processing within a single Hoppr stage
    """

    def __init__(
        self,
        stage_id: str,
        context: Context,
        plugin_ref_list: List[PluginRef],
        max_processes=3,
    ):
        self.stage_id = stage_id
        self.context = context
        self.plugin_ref_list = plugin_ref_list
        self.max_processes = max_processes
        self.plugins: List[HopprPlugin] = []
        self.results: Dict = {}

    def run(self) -> Result:
        """
        Run all processes for this stage
        """

        HopprProcessor.get_logger().info(  # pylint: disable=logging-fstring-interpolation
            f"---------- Beginning Stage {self.stage_id} ----------"
        )
        try:
            self.plugins = self._load_plugins()
        except (ModuleNotFoundError, HopprPluginError) as err:
            return Result.fail(str(err))

        result = Result.success()

        # Run each sub-stage (pre_stage_process, process_component, post_stage_process)
        # for all plugins (and, for process_component, for all components).
        # Each sub-stage must complete before the next can begin

        result.merge(self._run_all(HopprPlugin.pre_stage_process.__name__))
        result.merge(self._run_all(HopprPlugin.process_component.__name__))
        result.merge(self._run_all(HopprPlugin.post_stage_process.__name__))

        return result

    def _run_all(self, method_name: str):
        """
        Run the named method for all plugins.  If appropriate to the method, run it for
        all components for all plug-ins.
        """

        component_based_methods = [HopprPlugin.process_component.__name__]
        futures = []

        # Map to allow access to the arguments that went into a future call
        future_argument_map: Dict[Future, Tuple[HopprPlugin, Optional[Component]]] = {}

        with ProcessPoolExecutor(max_workers=self.max_processes) as executor:
            for plugin in self.plugins:
                if method_name in component_based_methods:

                    # Create one concurrent future object to run this method for each component

                    for component in self.context.consolidated_sbom.components or []:
                        future_proc = executor.submit(
                            _run_plugin, plugin, method_name, component
                        )
                        future_argument_map[future_proc] = (plugin, component)
                        futures.append(future_proc)
                else:

                    # Create a concurrent future object to run this method

                    future_proc = executor.submit(
                        _run_plugin, plugin, method_name, None
                    )
                    future_argument_map[future_proc] = (plugin, None)
                    futures.append(future_proc)

            # Save all the results, count failures and retries
            # Note: future.results() blocks until the process is complete

            failures = 0
            retries = 0
            for future_proc in futures:
                future_result = future_proc.result()
                if not future_result.is_skip():
                    plugin, comp = future_argument_map[future_proc]
                    self._save_result(method_name, plugin, future_result, comp)

                    if future_result.is_fail():
                        failures += 1
                    if future_result.is_retry():
                        retries += 1

        if failures + retries == 0:
            return Result.success()
        if failures == 0 and retries > 0:
            return Result.fail(f"{retries} '{method_name}' processes returned 'retry'")
        if failures > 0 and retries == 0:
            return Result.fail(f"{failures} '{method_name}' processes failed")
        return Result.fail(
            f"{failures} '{method_name}' processes failed, and {retries} returned 'retry'"
        )

    def _save_result(
        self,
        method_name: str,
        plugin: HopprPlugin,
        result: Result,
        comp: Optional[Component],
    ):
        """
        Store the results for later use
        """
        comp_string = None
        if comp is not None:
            comp_purl = PackageURL.from_string(comp.purl)
            if comp_purl is not None:
                comp_string = comp_purl.to_string()

        # If needed, create a new list for this method
        # Might need to expand this definition in the future to separate by plug-in

        if not method_name in self.results:
            self.results[method_name] = []

        self.results[method_name].append(
            (plugin.__class__.__name__, comp_string, result)
        )

    def _load_plugins(self) -> List[HopprPlugin]:
        """
        Create a list consisting of a single instance of each plug-in class used in this stage
        """
        plugin_list = []
        for plugin_ref in self.plugin_ref_list:
            plugin_inst = plugin_instance(
                plugin_ref.name, self.context, plugin_ref.config
            )
            plugin_list.append(plugin_inst)

        return plugin_list


class HopprProcessor:  # pylint: disable=too-few-public-methods
    """
    Run the Hoppr process
    """

    @staticmethod
    def _create_logger() -> logging.Logger:
        """
        Creates a logger to stderr
        """

        logger_name: str = f"HopprProcessor--{os.getpid()}"

        new_logger = logging.getLogger(logger_name)
        new_logger.setLevel(logging.INFO)

        formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )
        handler = logging.StreamHandler()
        handler.setFormatter(formatter)

        new_logger.addHandler(handler)
        return new_logger

    logger = _create_logger()

    def __init__(
        self, transfer: Transfer, manifest: Manifest, credentials: Credentials
    ) -> None:
        self.transfer = transfer
        self.manifest = manifest
        self.credentials = credentials
        self.context: Context
        self.stages: Dict = {}
        self.metadata_files: List[Path] = []

    @classmethod
    def get_logger(cls):
        """
        Returns the logger for this class
        """
        return cls.logger

    def _collect_file(self, file_name, target_dir) -> None:
        self.get_logger().info(  # pylint: disable=logging-fstring-interpolation
            f"Collecting metadata file {file_name}"
        )
        abs_path = os.path.abspath(file_name)

        target = target_dir + urllib.parse.quote_plus(abs_path)
        shutil.copyfile(file_name, target)

    def _collect_url(self, url, target_dir) -> None:
        self.get_logger().info(  # pylint: disable=logging-fstring-interpolation
            f"Collecting metadata from url {url}"
        )
        target = target_dir + urllib.parse.quote_plus(url)
        creds = None
        if self.context.credentials is not None:
            creds = self.context.credentials.find_credentials(url)

        download_file(url, target, creds=creds)

    def _collect_manifest_metadata(self, manifest: Manifest, target_dir) -> None:
        for (child_ref, child) in zip_longest(
            manifest.manifest_file_content.includes, manifest.children
        ):
            if child_ref.local is not None:
                self._collect_file(child_ref.local, target_dir)
            else:
                self._collect_url(child_ref.url, target_dir)

            self._collect_manifest_metadata(child, target_dir)

        for bom_ref in manifest.manifest_file_content.sbom_refs:
            if bom_ref.local is not None:
                self._collect_file(bom_ref.local, target_dir)
            else:
                self._collect_url(bom_ref.url, target_dir)

    def _collect_metadata(self):
        target_dir = f"{self.context.collect_root_dir}/generic/_metadata_/"
        os.makedirs(target_dir, exist_ok=True)

        with open(target_dir + "_run_data_", "w", encoding="utf-8") as rundata:
            rundata.write(f"Collection Start: {str(datetime.now())}\n")
            rundata.write(f"User:             {pwd.getpwuid(os.getuid())[0]}\n")
            rundata.write(f"Host FQDN:        {socket.getfqdn()}\n")

        for file_name in self.metadata_files:
            self._collect_file(file_name, target_dir)

        self._collect_manifest_metadata(self.context.manifest, target_dir)

    def _collect_bom(self) -> None:
        target_dir = f"{self.context.collect_root_dir}/generic/_metadata_/"
        os.makedirs(target_dir, exist_ok=True)

        with open(
            target_dir + "_consolidated_bom.json", "w", encoding="utf-8"
        ) as bom_data:
            bom_data.write(self.context.consolidated_sbom.json(by_alias=True, indent=2))

    def _summarize_results(self) -> int:
        """
        Summarize the results of a HopprProcess run
        """

        echo("\n========== Results Summary ==========")
        total_success_count = 0
        total_failure_count = 0
        for stage_id, stage in self.stages.items():
            echo(f"\nStage: {stage_id}")
            failure_list = "\n      Failure Summary:\n"
            for method_name, result_list in stage.results.items():
                echo(f"   {method_name}")
                result_count = len(result_list)
                failure_count = 0
                for plugin_name, comp_str, result in result_list:

                    # All retries should be handled internally by the plugins,
                    # So if a RETRY result is returned, that's a failure

                    if result.is_fail() or result.is_retry():
                        failure_count += 1
                        failure_list += f"         {plugin_name}: "
                        if comp_str is not None:
                            failure_list += f"Component: {comp_str}: "
                        failure_list += result.message + "\n"

                total_failure_count += failure_count
                total_success_count += result_count - failure_count
                echo(
                    f"      {result_count - failure_count} jobs succeeded, {failure_count} failed"
                )
                if failure_count > 0:
                    echo(failure_list)

        echo(
            f"\nGRAND TOTAL: {total_success_count} jobs succeeded, {total_failure_count} failed\n"
        )
        return total_failure_count

    def run(self) -> Result:
        """
        Run the Hoppr process executing each stage in turn
        """
        result = Result.success()

        with tempfile.TemporaryDirectory() as collection_root:

            self.context = Context(
                manifest=self.manifest,
                credentials=self.credentials,
                collect_root_dir=collection_root,
                consolidated_sbom=flatten_sboms(self.manifest),
            )

            self._collect_metadata()

            if self.transfer.content is not None:
                for stage_ref in self.transfer.content.stages:
                    stage = StageProcessor(
                        stage_ref.name, self.context, stage_ref.plugins
                    )
                    self.stages[stage_ref.name] = stage
                    result = stage.run()
                    if result.is_fail() or result.is_retry():
                        self.logger.error(  # pylint: disable=logging-fstring-interpolation
                            f"Stage {stage_ref.name} failed, processing terminated"
                        )
                        break

                    self._collect_bom()

        failed_jobs = self._summarize_results()
        if failed_jobs > 0:
            result.merge(Result.fail(f"{failed_jobs} failed during this execution"))

        return result



"""
Transfer business logic
"""

# Enable forward definitions
from __future__ import annotations

from hoppr import utils
from hoppr.types.transfer_file_content import TransferFileContent


class Transfer:
    """
    Transfer business logic class
    """

    def __init__(self) -> None:
        self.content = None

    @staticmethod
    def load_file(file) -> Transfer:
        """
        Creates a transfer object from a file
        """
        input_dict = utils.load_file(file)
        transfer = Transfer()
        transfer.populate(input_dict)

        return transfer

    def populate(self, input_dict):
        """
        Populates Transfer object with dictionary contents.
        """

        self.content = TransferFileContent(**input_dict)



"""
ManifestType dataclass and constructor
"""

from dataclasses import dataclass
from typing import Optional, Set


@dataclass
class CredentialRequiredService:
    """
    CredentialRequiredService
    """

    url: str
    user: Optional[str]
    user_env: Optional[str]
    pass_env: str

    def __init__(
        self, url: str = "", user: str = None, user_env: str = None, pass_env: str = ""
    ) -> None:
        self.url = url
        self.user = user
        self.user_env = user_env
        self.pass_env = pass_env

    def __hash__(self):
        """
        Hash must be defined to allow Repository objects to be included in a set
        """
        return hash(repr(self))


@dataclass
class CredentialMetadata:  # pylint: disable="duplicate-code"
    """
    CredentialMetadata Dataclass containing name, version, description
    """

    name: str
    version: str
    description: str

    def __init__(self, name: str, version: str, description: str) -> None:
        self.name = name
        self.version = version
        self.description = description


class CredentialFileContent:  # pylint: disable="too-few-public-methods"
    """
    Manifest data type class to construct ManifestType Object
    """

    schema_version: str
    kind: str
    metadata: CredentialMetadata
    credential_required_services: Set[CredentialRequiredService]

    def __init__(
        self,
        schemaVersion: str,  # pylint: disable="invalid-name"
        kind: str,
        metadata: dict,
        credential_required_services: Set[dict],
    ) -> None:
        self.schema_version = schemaVersion
        self.kind = kind
        self.metadata = CredentialMetadata(**metadata)
        self.credential_required_services = {
            CredentialRequiredService(**service)
            for service in credential_required_services
        }



"""
Collector plugin for pypi images
"""
from typing import Any, Dict, Optional

from packageurl import PackageURL  # type: ignore

from hoppr import __version__
from hoppr.base_plugins.collector import CollectorPlugin
from hoppr.base_plugins.hoppr import hoppr_rerunner
from hoppr.context import Context
from hoppr.result import Result
from hoppr.types.cred_object import CredObject


class CollectPypiPlugin(CollectorPlugin):
    """
    Collector plugin for pypi images
    """

    supported_purl_types = ["pip", "pypi"]
    required_commands = ["pip"]

    def get_version(self) -> str:  # pylint: disable=duplicate-code
        return __version__

    def __init__(self, context: Context, config: Optional[Dict] = None) -> None:
        super().__init__(context=context, config=config)
        if self.config is not None:
            if "pip_command" in self.config:
                self.required_commands = [self.config["pip_command"]]

    @hoppr_rerunner
    def collect(self, comp: Any, repo_url: str, creds: CredObject = None):
        """
        Copy a component to the local collection directory structure
        """
        purl = PackageURL.from_string(comp.purl)

        source_url = f"{repo_url}/simple"
        password_list = []
        if creds is not None:
            source_url = source_url.replace(
                "://", f"://{creds.username}:{creds.password}@", 1
            )
            password_list = [creds.password]

        target_dir = self.directory_for(
            purl.type, repo_url, subdir=f"{purl.name}_{purl.version}"
        )

        self.get_logger().info(  # pylint: disable=logging-fstring-interpolation
            f"Copying pypi artifact {purl.name}:{purl.version} from {repo_url} to {target_dir}"
        )

        command = [
            self.required_commands[0],
            "download",
            "--no-deps",
            f"{purl.name}=={purl.version}",
            "--index-url",
            source_url,
            "-d",
            target_dir,
            "--no-cache",
            "--timeout",
            "60",
            "--only-binary=:all:",
        ]

        got_binary = True

        result = self.run_command(command, password_list)
        if result.returncode != 0:
            self.get_logger().error(  # pylint: disable=logging-fstring-interpolation
                f"Failed to download {purl.name} version {purl.version} binary"
            )
            got_binary = False

        command[-1] = "--no-binary=:all:"

        result = self.run_command(command, password_list)
        if result.returncode != 0:
            self.get_logger().error(  # pylint: disable=logging-fstring-interpolation
                f"Failed to download {purl.name} version {purl.version} source"
            )
            if not got_binary:
                msg = (
                    f"Failed to download {purl.name} version {purl.version} "
                    + "as either whl or source."
                )
                return Result.retry(msg)

        return Result.success()



"""
Enumeration of supported purl types
"""

from enum import Enum


class PurlType(Enum):
    """
    Enumeration of supported purl types
    """

    DOCKER = "docker"
    GIT = "git"
    HELM = "helm"
    RPM = "rpm"
    PYPI = "pypi"
    MAVEN = "maven"
    GENERIC = "generic"
    APT = "apt"



"""
Manifest business logic
"""
from __future__ import annotations

from pathlib import Path
from typing import List, Optional

from typer import echo

from hoppr import utils
from hoppr.configs.credentials import Credentials
from hoppr.exceptions import HopprLoadDataError
from hoppr.types.cyclonedx_1_3 import (
    CyclonedxSoftwareBillOfMaterialSpecification as Bom_1_3,
)
from hoppr.types.cyclonedx_1_4 import (
    CyclonedxSoftwareBillOfMaterialsStandard as Bom_1_4,
)
from hoppr.types.manifest_file_content import ManifestFileContent, Repositories
from hoppr.types.purl_type import PurlType
from hoppr.utils import dedup_list

# Enable forward definitions


class Manifest:
    """
    Manifest business logic class
    """

    loaded_manifests: List[str] = []

    def __init__(self) -> None:
        self.manifest_file_content: ManifestFileContent
        self.sboms: List[Bom_1_4 | Bom_1_3] = []
        self.parent: Optional[Manifest] = None
        self.children: List["Manifest"] = []
        self.consolidated_repositories: Repositories

    @staticmethod
    def merge_repositories(
        first: Repositories,
        second: Repositories,
    ):
        """
        Add repos
        """
        combined: Repositories = {}
        for purl_type in PurlType:
            combined[purl_type] = dedup_list(first[purl_type] + second[purl_type])

        return combined

    @staticmethod
    def load_file(
        file: Path, credentials: Credentials = None, parent: Optional[Manifest] = None
    ) -> Manifest:
        """
        Creates a manifest object from a file
        """
        Manifest.loaded_manifests.append(str(file.resolve()))
        input_dict = utils.load_file(file)
        manifest = Manifest()
        manifest.populate(input_dict, credentials, parent)
        return manifest

    @staticmethod
    def load_url(
        url, credentials: Credentials = None, parent: Optional[Manifest] = None
    ) -> Manifest:
        """
        Creates a manifest object from a url
        """
        auth = None
        if credentials is not None:
            personal_cred = credentials.find_credentials(url)
            auth = (
                (personal_cred.username, personal_cred.password)
                if personal_cred
                else None
            )

        Manifest.loaded_manifests.append(url)
        input_dict = utils.load_url(url, auth)
        manifest = Manifest()
        manifest.populate(input_dict, credentials, parent)
        return manifest

    @staticmethod
    def load_manifest(
        manifest_location,
        credentials: Credentials = None,
        parent: Optional[Manifest] = None,
    ) -> Manifest:
        """
        Loads manifest
        """
        if manifest_location.url is not None:
            child_manifest = Manifest.load_url(
                manifest_location.url, credentials, parent
            )
        if manifest_location.local is not None:
            local_path = Path(manifest_location.local)
            child_manifest = Manifest.load_file(local_path, credentials, parent)

        return child_manifest

    @staticmethod
    def load_sbom(sbom_location, credentials: Credentials = None) -> Bom_1_4 | Bom_1_3:
        """
        Loads SBOM from urls and files
        """

        if sbom_location.local is not None:
            sbom_file_object = utils.load_file(Path(sbom_location.local))
        elif sbom_location.url is not None:
            sbom_file_object = utils.load_url(sbom_location.url, credentials)
        else:
            raise HopprLoadDataError(f"Unsupported SBOM Location {sbom_location}")

        spec_version = sbom_file_object.get("specVersion", "")

        if spec_version == "1.4":
            return Bom_1_4(**sbom_file_object)
        if spec_version == "1.3":
            return Bom_1_3(**sbom_file_object)
        raise HopprLoadDataError(
            f"{sbom_location} is an unknown spec version ({spec_version})"
        )

    def build_repository_search(self) -> None:
        """
        Builds the computed repository search sequence for a parent and child manifest
        """
        if self.parent is None:
            self.consolidated_repositories = self.manifest_file_content.repositories
        else:
            self.consolidated_repositories = self.merge_repositories(
                self.parent.consolidated_repositories,
                self.manifest_file_content.repositories,
            )

    def populate(
        self,
        input_dict,
        credentials: Optional[Credentials] = None,
        parent: Optional[Manifest] = None,
    ):
        """
        Populates Manifest object with dictionary contents.
        """

        self.manifest_file_content = ManifestFileContent(**input_dict)
        self.parent = parent

        if self.manifest_file_content is not None:
            for sbom_ref in self.manifest_file_content.sbom_refs:
                self.sboms.append(Manifest.load_sbom(sbom_ref))

            self.build_repository_search()

            for include in self.manifest_file_content.includes:
                if (
                    include.url or str(Path(str(include.local)).resolve())
                ) in Manifest.loaded_manifests:
                    echo(
                        f"WARNING: Manifest file '{include.url or include.local}' "
                        + "has already been loaded.  Subsequent load requests ignored."
                    )
                else:
                    child = self.load_manifest(include, credentials, self)
                    self.children.append(child)



"""
Framework for manipulating bundles for airgapped transfers.
"""
import sys
from pathlib import Path
from platform import python_version
from typing import List

from typer import Option, Typer, echo

from hoppr import __version__
from hoppr.configs.credentials import Credentials
from hoppr.configs.manifest import Manifest
from hoppr.configs.transfer import Transfer
from hoppr.processor import HopprProcessor

app = Typer()


@app.command()
def bundle(
    manifest_file: Path,
    auth_config_file: Path = Option(
        None,
        help="Specify authorization config for services",
        envvar="HOPPR_AUTH_CONFIG",
    ),
    transfer_config_file: Path = Option(
        ...,
        help="Specify transfer config",
        envvar="HOPPR_TRANSFER_CONFIG",
    ),
):
    """
    Run the stages specified in the transfer config file on the
    content specified in the manifest
    """

    metadata_files = [manifest_file, transfer_config_file]

    manifest = Manifest.load_file(manifest_file)
    credentials = None
    if auth_config_file is not None:
        credentials = Credentials.load_file(auth_config_file)
        metadata_files.append(auth_config_file)

    transfer = Transfer.load_file(transfer_config_file)

    processor = HopprProcessor(transfer, manifest, credentials)
    processor.metadata_files = metadata_files

    result = processor.run()
    if result.is_fail():
        sys.exit(1)


@app.command()
def version():
    """
    Print version information for `hoppr`
    """
    echo(f"Hoppr Framework Version: {__version__}")
    echo(f"Python Version: {python_version()}")


@app.command()
def validate(
    input_files: List[Path],
    credentials_config_file: Path = Option(
        None,
        help="Specify authorization config for services",
        envvar="HOPPR_CREDS_CONFIG",
    ),
    transfer_config_file: Path = Option(
        None,
        help="Specify transfer config for services",
        envvar="HOPPR_TRANSFER_CONFIG",
    ),
):
    """
    Validate multiple manifest files for schema errors.
    """

    cred_config = None
    transfer_config = None  # pylint: disable="unused-variable"
    manifests = []  # pylint: disable="unused-variable"

    if credentials_config_file is not None:
        cred_config = Credentials.load_file(credentials_config_file)
    if transfer_config_file is not None:
        transfer_config = Transfer.load_file(transfer_config_file)

    manifests = [Manifest.load_file(file, cred_config) for file in input_files]


if __name__ == "__main__":  # pragma: no cover
    app()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['hoppr',
 'hoppr.base_plugins',
 'hoppr.configs',
 'hoppr.core_plugins',
 'hoppr.types',
 'hoppr.types.cyclonedx_1_3',
 'hoppr.types.cyclonedx_1_4']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'click==8.1.3',
 'cyclonedx-python-lib>=2.0.0,<3.0.0',
 'jsonschema>=4.4.0,<5.0.0',
 'packageurl-python>=0.9.9,<0.10.0',
 'pydantic[email]>=1.9.0,<2.0.0',
 'requests>=2.27.1,<3.0.0',
 'typer>=0.4.0,<0.5.0',
 'types-PyYAML>=6.0.5,<7.0.0',
 'urllib3>=1.26.9,<2.0.0']

entry_points = \
{'console_scripts': ['hopctl = hoppr.main:app']}

setup_kwargs = {
    'name': 'hoppr',
    'version': '0.18.2',
    'description': 'A tool for defining, verifying, and transferring software dependencies between environments.',
    'long_description': None,
    'author': 'LMCO Open Source',
    'author_email': 'open.source@lmco.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/lmco/hoppr/hoppr',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



"""
Collector plugin for git repositories
"""
import pathlib
import re
from typing import Any, Dict, Optional

from packageurl import PackageURL  # type: ignore

from hoppr import __version__
from hoppr.base_plugins.collector import CollectorPlugin
from hoppr.base_plugins.hoppr import hoppr_rerunner
from hoppr.context import Context
from hoppr.result import Result
from hoppr.types.cred_object import CredObject


class CollectGitPlugin(CollectorPlugin):
    """
    Class to copy git repositories
    """

    supported_purl_types = ["git"]
    required_commands = ["git"]

    def get_version(self) -> str:  # pylint: disable=duplicate-code
        return __version__

    def __init__(self, context: Context, config: Optional[Dict] = None) -> None:
        super().__init__(context=context, config=config)
        if self.config is not None:
            if "git_command" in self.config:
                self.required_commands = [self.config["git_command"]]

    @hoppr_rerunner
    def collect(self, comp: Any, repo_url: str, creds: CredObject = None):
        """
        Collect git repository
        """
        purl = PackageURL.from_string(comp.purl)

        self.get_logger().info(  # pylint: disable=logging-fstring-interpolation
            f"Copying git repository from {repo_url}"
        )

        source_url = f"{repo_url}/{purl.namespace}/{purl.name}"
        target_dir = self.directory_for(
            purl.type, repo_url, subdir=f"{purl.namespace}/{purl.name}"
        )

        result = self.git_clone(target_dir, source_url, creds)
        if not result.is_success():
            return result

        result = self.git_update(target_dir, purl.name)
        if not result.is_success():
            return result

        return Result.success()

    def git_clone(self, tmp_dir, source_url, source_creds):
        """Git clone"""

        git_src = source_url
        password_list = []
        if re.match("^https?://", git_src) and source_creds is not None:
            git_src = git_src.replace("://", f"://{source_creds.username}@", 1)
            if source_creds.password:
                git_src = git_src.replace("@", f":{source_creds.password}@")
                password_list = [source_creds.password]

        if git_src.startswith("ssh://") and source_creds is not None:
            git_src = git_src.replace("ssh://", f"ssh://{source_creds.username}@")

        if not git_src.endswith(".git"):
            git_src += ".git"

        result = self.run_command(
            [self.required_commands[0], "clone", "--mirror", git_src],
            password_list,
            cwd=tmp_dir,
        )
        if result.returncode != 0:
            msg = f"Failed to clone {source_url}"
            self.get_logger().error(msg)
            return Result.retry(msg)

        return Result.success()

    def git_update(self, tmp_dir, name_git):
        """Git update-server-info"""

        if not name_git.endswith(".git"):
            name_git += ".git"

        repo_dir = pathlib.Path(tmp_dir) / name_git

        # Make the clone usable as a remote
        result = self.run_command(
            [self.required_commands[0], "update-server-info"],
            cwd=repo_dir,
        )
        if result.returncode != 0:
            msg = "Failed to make the clone usable as a remote"
            self.get_logger().error(msg)
            return Result.retry(msg)

        return Result.success()
